from pyfyre.styles import Style
from pyfyre.states import BaseState
from pyfyre.nodes.base import Node, Element
from typing import Any, Callable, Optional, Dict, List


class TextInput(Element):
    """Represents an HTML ``<input>`` or ``<textarea>`` depending on the ``multiline`` argument.

    Args:
        placeholder: Placeholder text of this input.
        multiline: Whether this text input is multiline or not.
    """

    def __init__(
        self,
        children: Optional[Callable[[], List[Node]]] = None,
        *,
        placeholder: str = "",
        multiline: bool = False,
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        attrs = attrs or {}
        attrs["placeholder"] = placeholder

        super().__init__(
            "textarea" if multiline else "input",
            children,
            styles=styles,
            states=states,
            attrs=attrs,
        )

    @property
    def value(self) -> str:
        """Value of this text input."""
        return str(self.dom.value)

    def set_value(self, value: Any) -> None:
        """Set the value of this text input."""
        self.dom.value = str(value)
        self.update_dom()



import sys
from pyfyre.styles import Style
from pyfyre.states import BaseState
from browser import timer
from browser import DOMEvent
from pyfyre.nodes import Node, Element
from typing import Optional, Dict, List, Callable


class ListBuilder(Element):
    """A scrollable list of nodes that are created on demand.

    This element is appropriate for lists with a large (or infinite) number of children
    because the builder is called only for those children that are actually visible.

    Args:
        item_builder: Builder for a child of this list.
            Called only with indices greater than or equal to zero and less than ``count``.
            None can be returned in the builder to skip the current index.
        count: Number of times this list will attempt to build a child.
            If None, infinite attempts will be made.
        max_height: Maximum height of this element.
            A ``max-height`` CSS property is required to make a scrollable element.
        render_batch: Number of children built at the same time when the user scrolls the list.
        render_interval: Time interval in seconds between each batch of children build.

    Attributes:
        item_builder (Callable[[int], Optional[~pyfyre.nodes.Node]]):
            Builder for a child of this list.
            Called only with indices greater than or equal to zero and less than ``count``.
            None can be returned in the builder to skip the current index.
        count (Optional[int]): Number of times this list will attempt to build a child.
            If None, infinite attempts will be made.
        max_height (str): Maximum height of this element.
            A ``max-height`` CSS property is required to make a scrollable element.
        render_batch (int): Number of children built at the same time when the user scrolls the list.
        render_interval (float): Time interval in seconds between each batch of children build.
        index (int): Current index of this list builder.
    """

    def __init__(
        self,
        item_builder: Callable[[int], Optional[Node]],
        *,
        count: Optional[int] = None,
        max_height: str = "300px",
        render_batch: int = 10,
        render_interval: float = 0,
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        styles = styles or []
        self.item_builder = item_builder
        self.count = count
        self.max_height = max_height
        self.render_batch = render_batch
        self.render_interval = render_interval

        self.index = 0
        self._rendered_children: List[Node] = []

        super().__init__(
            "div",
            lambda: self._rendered_children,
            attrs=attrs,
            styles=[
                Style(
                    overflow_y="scroll",
                    overflow_wrap="break-word",
                    max_height=max_height,
                )
            ]
            + styles,
            states=states,
        )

        def render_nodes(event: DOMEvent) -> None:
            el = event.target
            if el.scrollHeight - el.scrollTop - el.clientHeight < 1:
                rendered = self.render_next_children()
                if rendered:
                    render_nodes(event)

        self.add_event_listener("scroll", render_nodes)

    def render_next_children(self) -> bool:
        """Build the next batch of children of this list.

        Returns:
            True if a new child is built. Otherwise, False.
        """
        prev_index = self.index

        for _ in range(self.render_batch):
            if self.count is not None:
                if self.index >= self.count:
                    break

            try:
                child = self.item_builder(self.index)
            except Exception:
                error_children = self.on_build_error(*sys.exc_info())
                child = Element("div", lambda: error_children)

            if child is not None:
                self._rendered_children.append(child)

            self.index += 1

        if self.index > prev_index:
            self.update_dom()
            return True

        return False

    def build_children(self, *, propagate: bool = True) -> None:
        super().build_children(propagate=propagate)

        def render_next_children() -> None:
            if self.dom.scrollHeight == self.dom.clientHeight:
                self.render_next_children()
            else:
                timer.clear_interval(interval)

        interval = timer.set_interval(render_next_children, self.render_interval)



from typing import Dict, List, Any

# When building your app, PyFyre creates HTML files based on these routes,
# so that when a user visits your website,
# the corresponding page of the route is returned.
# Once the initial page has been loaded,
# PyFyre will take over the routing as a single-page application.
#
# -> A user visits "https://your-pyfyre-app.com/about".
# -> The server returns the `about/index.html`.
# -> PyFyre takes over the page routing
#    by preventing further navigation to internal links.
# -> The user clicks an internal link "https://your-pyfyre-app.com/home".
# -> PyFyre prevents the navigation and modifies the DOM
#    to render the corresponding view of the route "/home".
#
# Example:
# {
# 	"/": {
# 		"title": "A PyFyre App",
# 		"icon": "/favicon.ico",
# 		"head": ['<link rel="stylesheet" href="/style.css" />']
# 	}
# }
ROUTES: Dict[str, Dict[str, Any]] = {
    "/": {
        "title": "A PyFyre App",
        "icon": "/favicon.ico",
        "head": ['<link rel="stylesheet" href="/style.css" />'],
    }
}

# When building your app, PyFyre makes your CPython packages installed by pip
# usable for the web. Just add the name of the pip package in the list.
#
# !!! Note that it is not possible to import every CPython package
# due to Brython limitations. Read the Brython documentation:
# https://www.brython.info/static_doc/en/faq.html#:~:text=Q%20%3A%20can%20I%20import%20all%20the%20modules%20/%20packages%20that%20run%20with%20CPython%20%3F
#
# Example:
# [
# 	"text_generator",
# 	"random_string",
# 	"url64"
# ]
PYTHON_DEPENDENCIES: List[str] = []



from enum import Enum
from typing import Callable
from browser import DOMEvent, window


class PyFyreEventType(Enum):
    """PyFyre event types."""

    pyfyreload = "pyfyreload"


def window_event_listener(
    event_type: str,
) -> Callable[[Callable[[DOMEvent], None]], Callable[[DOMEvent], None]]:
    """A decorator function.

    Calls the decorated function whenever the specified ``event_type`` is delivered to the window.

    **Example:**

    .. code-block:: python

        @window_event_listener("error")
        def onerror(event: DOMEvent) -> None:
            ...
    """

    def window_event_listener_decorator(
        callback: Callable[[DOMEvent], None]
    ) -> Callable[[DOMEvent], None]:
        window.bind(event_type, callback)
        return callback

    return window_event_listener_decorator



from types import TracebackType
from abc import ABC, abstractmethod
from typing import Optional, Dict, List, Type
from pyfyre.styles import Style
from pyfyre.states import BaseState
from pyfyre.nodes import Node, Element, FutureElement


class Widget(Element, ABC):
    """User-defined ``Element`` class.

    Inherit this class if you wish to define your own ``Element`` class.

    **Example:**

    .. code-block:: python

        class MyElement(Widget):
            def build(self) -> list[Node]:
                return [Element("p", lambda: [Text("Hello, World!")])]
    """

    def __init__(
        self,
        *,
        tag_name: str = "div",
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        super().__init__(
            tag_name, self.build, styles=styles, states=states, attrs=attrs
        )

    @abstractmethod
    def build(self) -> List[Node]:
        """Builder for the children nodes of this element.

        This is an abstract method so you must override this and define its implementation.
        """
        raise NotImplementedError

    def on_build_error(
        self,
        exc_type: Type[Exception],
        exc_value: Exception,
        exc_traceback: TracebackType,
    ) -> List[Node]:
        """Error handler for the ``build`` method.

        This replaces the ``build`` method as the children builder for this element
        if an error occurred inside the ``build`` method.

        You may override this method to customize your error message.

        Args:
            exc_type: Type of the exception.
            exc_value: The exception itself.
            exc_traceback: Traceback object which typically encapsulates the
                call stack at the point where the exception last occurred.

        Returns:
            By default, a simple error message on production environment
            and a traceback message on development environment.
        """
        return super().on_build_error(exc_type, exc_value, exc_traceback)


class FutureWidget(FutureElement, ABC):
    """User-defined ``FutureElement`` class.

    Inherit this class if you wish to define your own ``FutureElement`` class.

    **Example:**

    .. code-block:: python

        class MyFutureElement(FutureWidget):
            async def build(self) -> list[Node]:
                return await ...  # asynchronous operation
    """

    def __init__(
        self,
        *,
        tag_name: str = "div",
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        super().__init__(
            tag_name, self.build, styles=styles, states=states, attrs=attrs
        )

    @abstractmethod
    async def build(self) -> List[Node]:
        """Asynchronous builder for the children nodes of this Future element.

        This is an abstract method so you must override this and define its implementation.

        Error handling for the ``build`` method is currently not yet supported.
        """
        raise NotImplementedError

    def while_no_result(self) -> List[Node]:
        """Initial children builder for this element
        while the result of the Future is not yet ready.

        You may override this method to customize your loading message.

        Returns:
            By default, a loading animation.
        """
        return super().while_no_result()



from livereload import Server
from pyfyre_cli.web_app import create_static_flask_app
from pyfyre_cli.build_app import build_app, bundle_scripts, create_pages


def run_app(port: str, hostname: str) -> None:
    """Builds the app in the `_pyfyre` directory and starts a live server hosting that directory.
    The live server watches for changes in the `public` and `src` directories,
    as well as the `settings.py` and `template.html` files.
    When the live server detects changes, it will rebuild the app.
    """
    build_app()

    app = create_static_flask_app("_pyfyre")
    app.debug = True

    server = Server(app.wsgi_app)
    server.watch("public/", build_app)
    server.watch("src/", lambda: bundle_scripts(production=False))
    server.watch("settings.py", build_app)
    server.watch("template.html", lambda: create_pages(production=False))
    server.serve(port=port, host=hostname)



from typing import Any


class EmptyObject:
    def __init__(self, class_name: str) -> None:
        self.class_name = class_name
        self.__qualname__ = class_name

    def __getattr__(self, attr: str) -> Any:
        return EmptyObject(attr)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return EmptyObject("")

    def __getitem__(self, key: str) -> Any:
        return EmptyObject(key)

    def __str__(self) -> str:
        return self.class_name

    def __add__(self, other: str) -> str:
        return str(self) + other

    def __radd__(self, other: str) -> str:
        return other + str(self)


import_js: Any = EmptyObject("import_js")
py2js: Any = EmptyObject("py2js")
this: Any = EmptyObject("this")
Date: Any = EmptyObject("Date")
JSON: Any = EmptyObject("JSON")
Math: Any = EmptyObject("Math")
NULL: Any = EmptyObject("NULL")
Number: Any = EmptyObject("Number")
RegExp: Any = EmptyObject("RegExp")
String: Any = EmptyObject("String")
UNDEFINED: Any = EmptyObject("UNDEFINED")



from typing import List, Optional
from pyfyre.utils import EventMixin


class Style(EventMixin):
    """Provides styling to elements.

    Args:
        **props: CSS properties but the hyphens (-) are replaced by underscores (_).
            The values are unchanged.

    Attributes:
        props (Dict[str, str]): Packed form of the ``**props`` argument.

    **Example:**

    .. code-block:: python

        theme = Style(background_color="gray")

        Element(
            "p",
            lambda: [Text("PyFyre is amazing!")],
            styles=[theme]  # Sets the background color of the element to gray.
        )

        theme["background_color"]  # Returns "gray".
        theme["background_color"] = "black"  # Change the background color of the element to black.
        del theme["background_color"]  # Remove the background color styling of the element.
    """

    @classmethod
    def from_styles(cls, styles: List["Style"], *, dynamic: bool = True) -> "Style":
        """Construct a ``Style`` object from multiple ``Style`` objects.
        The styles are evaluated in order from left to right.

        Args:
            dynamic: If True, the constructed ``Style`` object will listen
                to the changes of the ``styles``.
                That means if one of the ``styles`` is updated,
                the constructed ``Style`` object will also get updated.
                If False, the constructed ``Style`` object will not
                listen to the changes of the ``styles``.
        """

        s = Style()
        props = {}

        def update_style() -> None:
            s.props = Style.from_styles(styles, dynamic=False).props.copy()
            s.call_listeners()

        for style in styles:
            for prop, value in style.props.items():
                props[prop] = value

            if dynamic:
                style.add_listener(update_style)

        s.props = props
        return s

    def __init__(self, **props: str) -> None:
        super().__init__()
        self.props = props

    def __getitem__(self, prop: str) -> Optional[str]:
        return self.props.get(prop)

    def __setitem__(self, prop: str, value: str) -> None:
        self.props[prop] = value
        self.call_listeners()

    def __delitem__(self, prop: str) -> None:
        del self.props[prop]
        self.call_listeners()

    def update(self, **props: str) -> None:
        """Update the properties of this style.

        Args:
            **props: CSS properties but the hyphens (-) are replaced by underscores (_).
                The values are unchanged.
        """
        for prop, value in props.items():
            self.props[prop] = value

        self.call_listeners()

    def set(self, **props: str) -> None:
        """Overwrite the properties of this style.

        Args:
            **props: CSS properties but the hyphens (-) are replaced by underscores (_).
                The values are unchanged.
        """
        self.props = props
        self.call_listeners()

    def css(self) -> str:
        """CSS representation of this object."""
        return "; ".join(
            [f"{prop.replace('_', '-')}: {value}" for prop, value in self.props.items()]
        )



import os
import sys
import shutil
from typing import Iterator
from contextlib import contextmanager


@contextmanager
def in_path(path: str, *, append_to_sys_path: bool = False) -> Iterator[str]:
    original_path = os.getcwd()
    abspath = os.path.abspath(path)

    try:
        if append_to_sys_path:
            sys.path.append(abspath)

        os.chdir(abspath)
        yield abspath
    finally:
        if append_to_sys_path:
            sys.path.remove(abspath)

        os.chdir(original_path)


def empty_directory(path: str) -> None:
    for f in os.listdir(path):
        p = os.path.join(path, f)

        try:
            shutil.rmtree(p)
        except NotADirectoryError:
            os.remove(p)



from typing import Any


class EmptyObject:
    def __init__(self, class_name: str) -> None:
        self.class_name = class_name
        self.__qualname__ = class_name

    def __getattr__(self, attr: str) -> Any:
        return EmptyObject(attr)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return EmptyObject("")

    def __getitem__(self, key: str) -> Any:
        return EmptyObject(key)

    def __str__(self) -> str:
        return self.class_name

    def __add__(self, other: str) -> str:
        return str(self) + other

    def __radd__(self, other: str) -> str:
        return other + str(self)


alert: Any = EmptyObject("alert")
bind: Any = EmptyObject("bind")
confirm: Any = EmptyObject("confirm")
console: Any = EmptyObject("console")
document: Any = EmptyObject("document")
DOMEvent: Any = EmptyObject("DOMEvent")
DOMNode: Any = EmptyObject("DOMNode")
load: Any = EmptyObject("load")
prompt: Any = EmptyObject("prompt")
run_script: Any = EmptyObject("run_script")
window: Any = EmptyObject("window")
aio: Any = EmptyObject("aio")
ajax: Any = EmptyObject("ajax")
html: Any = EmptyObject("html")
local_storage: Any = EmptyObject("local_storage")
markdown: Any = EmptyObject("markdown")
object_storage: Any = EmptyObject("object_storage")
session_storage: Any = EmptyObject("session_storage")
svg: Any = EmptyObject("svg")
template: Any = EmptyObject("template")
timer: Any = EmptyObject("timer")
webcomponent: Any = EmptyObject("webcomponent")
websocket: Any = EmptyObject("websocket")
worker: Any = EmptyObject("worker")
widgets: Any = EmptyObject("widgets")



from browser import aio
from pyfyre.styles import Style
from pyfyre.states import BaseState
from pyfyre.nodes import Node, Element
from typing import Dict, List, Optional, Callable, Awaitable
from pyfyre.exceptions import FutureNoResult, FutureCancelled, FutureAlreadyDone


class FutureElement(Element):
    """Just like an ``Element``, but the ``children`` returns an awaitable.
    That means the children builder of this element is asynchronous.

    This class’s implementation is similar to ``asyncio.Future``.
    """

    def __init__(
        self,
        tag_name: str,
        children: Callable[[], Awaitable[List[Node]]],
        *,
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        self._is_done = False
        self._is_cancelled = False
        self._result: Optional[List[Node]] = None
        self._exception: Optional[Exception] = None
        self._done_callbacks: List[Callable[["FutureElement"], None]] = []

        def process_result() -> List[Node]:
            try:
                return self.result()
            except FutureNoResult:
                return self.while_no_result()

        super().__init__(
            tag_name,
            process_result,
            styles=styles,
            states=states,
            attrs=attrs,
        )

        async def build_children(children: Callable[[], Awaitable[List[Node]]]) -> None:
            res = await children()

            try:
                self.set_result(res)
                self.update_dom()
            except FutureAlreadyDone:
                pass

        aio.run(build_children(children))

    def while_no_result(self) -> List[Node]:
        """:meta private:"""
        from pyfyre.presets.loading import Loading

        return [Loading()]

    def result(self) -> List[Node]:
        """Return the result of the Future.

        If the Future is `done` and has a result set by the ``set_result()`` method,
        the result value is returned.

        Returns:
            The result will be the children nodes of this element.

        Raises:
            Exception: If the Future is `done` and has an exception set by the
                ``set_exception()`` method, this method raises the exception.
            FutureCancelled: If the Future has been `cancelled`.
            FutureNoResult: If the Future’s result isn’t yet available.
        """
        if self.is_done():
            if self._result is not None:
                return self._result

            if self._exception is not None:
                raise self._exception

        if self.is_cancelled():
            raise FutureCancelled()

        raise FutureNoResult()

    def _mark_as_done(self) -> None:
        """Mark the Future as done and call the callbacks."""
        self._is_done = True

        for callback in self._done_callbacks:
            callback(self)

    def set_result(self, result: List[Node]) -> None:
        """Mark the Future as `done` and set its result.

        Raises:
            FutureAlreadyDone: If the Future is already `done`.
        """
        if self.is_done():
            raise FutureAlreadyDone()

        self._result = result
        self._mark_as_done()

    def set_exception(self, exception: Exception) -> None:
        """Mark the Future as `done` and set an exception.

        Raises:
            FutureAlreadyDone: If the Future is already `done`.
        """
        if self.is_done():
            raise FutureAlreadyDone()

        self._exception = exception
        self._mark_as_done()

    def is_done(self) -> bool:
        """A Future is `done` if it was `cancelled` or if it has a result or
        an exception set with ``set_result()`` or ``set_exception()`` calls.

        Returns:
            True if the Future is `done`.
        """
        return self.is_cancelled() or self._is_done

    def is_cancelled(self) -> bool:
        """
        Returns:
            True if the Future was `cancelled`.
        """
        return self._is_cancelled

    def add_done_callback(self, callback: Callable[["FutureElement"], None]) -> None:
        """Add a callback to be run when the Future is `done`.

        The ``callback`` is called with the Future object as its only argument.

        If the Future is already `done` when this method is called,
        the ``callback`` is called immediately.
        """
        self._done_callbacks.append(callback)

        if self.is_done():
            callback(self)

    def remove_done_callback(self, callback: Callable[["FutureElement"], None]) -> int:
        """Remove the ``callback`` from the callbacks list.

        Returns:
            The number of callbacks removed, which is typically 1,
            unless the ``callback`` was added more than once.
        """
        remaining_callbacks = []
        removed = 0

        for c in self._done_callbacks:
            if c == callback:
                removed += 1
            else:
                remaining_callbacks.append(c)

        self._done_callbacks = remaining_callbacks
        return removed

    def cancel(self) -> bool:
        """Cancel the Future and call the callbacks.

        Returns:
            False, if the Future is already `done` or `cancelled`.
            Otherwise, change the Future’s state to `cancelled`,
            call the callbacks, and return True.
        """
        if self.is_done():
            return False

        self._is_cancelled = True

        for callback in self._done_callbacks:
            callback(self)

        return True

    def exception(self) -> Optional[Exception]:
        """Return the exception that was set on this Future.

        Returns:
            The exception (or None if no exception was set), only if the Future is `done`.

        Raises:
            FutureCancelled: If the Future has been `cancelled`.
            FutureNoResult: If the Future isn’t `done` yet.
        """
        if self.is_done():
            return self._exception

        if self.is_cancelled():
            raise FutureCancelled()

        raise FutureNoResult()



from typing import List
from pyfyre import render, Style
from pyfyre.nodes import *


class App(Widget):
    def build(self) -> List[Node]:
        return [
            Element(
                "h1",
                lambda: [Text("Welcome to PyFyre 🐍🔥")],
                styles=[Style(font_family="Sans")],
            ),
            Element(
                "p",
                lambda: [
                    Text("Get started by modifying the "),
                    Element("code", lambda: [Text("src/index.py")]),
                    Text("."),
                ],
            ),
            Element(
                "p",
                lambda: [
                    Text("Learn PyFyre by reading the "),
                    Link(
                        "https://pyfyre-docs.netlify.app/",
                        lambda: [Text("documentation")],
                    ),
                    Text(". It is also advisable to learn "),
                    Link("https://www.brython.info/", lambda: [Text("Brython")]),
                    Text(" alongside PyFyre as it is built on top of Brython."),
                ],
            ),
            Element("h2", lambda: [Text("Deployment")]),
            Element(
                "p",
                lambda: [
                    Text("Run "),
                    Element("code", lambda: [Text("pyfyre build")]),
                    Text(". The build files are saved in the "),
                    Element("code", lambda: [Text("build")]),
                    Text(
                        " directory so you can just serve or deploy that directory to the web."
                    ),
                ],
            ),
            Element("h2", lambda: [Text("Links")]),
            Element(
                "ul",
                lambda: [
                    Element(
                        "li",
                        lambda: [
                            Link(
                                "https://pypi.org/project/pyfyre",
                                lambda: [Text("PyPI")],
                            )
                        ],
                    ),
                    Element(
                        "li",
                        lambda: [
                            Link(
                                "https://github.com/pyfyre/pyfyre",
                                lambda: [Text("Repository")],
                            )
                        ],
                    ),
                    Element(
                        "li",
                        lambda: [
                            Link(
                                "https://pyfyre-docs.netlify.app",
                                lambda: [Text("Documentation")],
                            )
                        ],
                    ),
                    Element(
                        "li",
                        lambda: [
                            Link(
                                "https://www.facebook.com/pyfyreframework",
                                lambda: [Text("Facebook Page")],
                            )
                        ],
                    ),
                    Element(
                        "li",
                        lambda: [
                            Link(
                                "https://discord.gg/YzEDuqhgZJ",
                                lambda: [Text("Discord Server")],
                            )
                        ],
                    ),
                ],
            ),
        ]


render({"/": lambda: App()})



import pyfyre
from abc import ABC
from typing import TypeVar, Generic, Optional
from browser import window
from pyfyre.utils import EventMixin

T = TypeVar("T")


class BaseState(ABC, EventMixin):
    """Abstract base class for state dependencies used for rebuilding nodes."""


class State(Generic[T], BaseState):
    """An object which stores a ``value`` of generic type ``T``.

    This object is used as a state dependency for nodes.

    **Example:**

    .. code-block:: python

        state = State(0)
        state = State[int](0)  # With generic type.

        Element(
            "p",
            lambda: [Text(f"Count: {state.value}")],  # It uses the value of the ``state``.
            states=[state]  # Set the ``state`` as a state dependency for this element.
        )

        state.set_value(1)  # The element will rebuild since it is depending on the ``state``.
    """

    def __init__(self, value: T) -> None:
        super().__init__()
        self._value = value

    @property
    def value(self) -> T:
        """The value of this state object."""
        return self._value

    def set_value(self, value: T) -> None:
        """Set the value and notify the dependers of this state object.

        If the ``value`` is the same, nothing will happen.
        """
        if self._value != value:
            self._value = value
            self.call_listeners()


class EventState(BaseState):
    """A state which updates whenever the specified ``event_type`` is delivered to the ``target``.
    If the ``target`` is None, will listen to the window instead.

    This object is used as a state dependency for elements.

    **Example:**

    .. code-block:: python

        Element(
            "p",
            lambda: [Text()],
            # The element will rebuild every time the window is scrolled.
            states=[EventState("scroll")]
        )
    """

    def __init__(
        self, event_type: str, target: "Optional[pyfyre.nodes.Element]" = None
    ) -> None:
        super().__init__()

        if target is not None:
            target.add_event_listener(event_type, lambda ev: self.call_listeners())
        else:
            window.bind(event_type, lambda ev: self.call_listeners())


class MediaQuery(BaseState):
    """A state which updates whenever the media query match has changed its state.

    This is similar to JavaScript ``window.matchMedia``.

    This object is used as a state dependency for elements.

    Args:
        media_query_string: Specifies the media query string to match.

    **Example:**

    .. code-block:: python

        Element(
            "p",
            lambda: [Text()],
            # The element will rebuild every time the media query matches.
            states=[MediaQuery("(max-width: 700px)")]
        )
    """

    def __init__(self, media_query_string: str) -> None:
        super().__init__()
        self._media_query = window.matchMedia(media_query_string)
        self._media_query.addListener(lambda mq: self.call_listeners())

    @property
    def matches(self) -> bool:
        """True if the window currently matches the media query, False otherwise."""
        return bool(self._media_query.matches)



import os
import shutil
from pyfyre_cli.utils import in_path, empty_directory


def create_app(app_name: str, app_dir: str, *, dev_mode: bool = False) -> None:
    if not dev_mode:
        print(f"Creating your PyFyre project '{app_name}'...")

    project_path = os.path.join(app_dir, app_name)

    if os.path.isdir(project_path):
        prompt = (
            "y"
            if dev_mode
            else input(
                f"Project '{app_name}' already exists.\nWant to overwrite "
                f"the directory? (y or n): "
            ).lower()
        )

        if prompt == "y":
            empty_directory(project_path)
        else:
            print("Aborting...")
            return

    with in_path(os.path.dirname(os.path.abspath(__file__))) as path:
        shutil.copytree(os.path.join(path, "user"), project_path, dirs_exist_ok=True)

    if not dev_mode:
        print("Project created successfully.")



import os
from flask import Flask, send_from_directory, Response


def create_static_flask_app(directory: str) -> Flask:
    app = Flask(__name__, root_path=os.path.abspath(directory))

    @app.route("/", defaults={"filename": "index.html"})
    @app.route("/<path:filename>")
    def index(filename: str) -> Response:
        target = os.path.join(directory, *filename.split("/"))

        if os.path.isfile(target):
            return send_from_directory(".", filename)

        return send_from_directory(".", os.path.join(filename, "index.html"))

    return app



from settings import ROUTES
from typing import Any, Dict, Callable, Optional
from browser import document, window, DOMEvent
from pyfyre.events import window_event_listener
from pyfyre.nodes import Node, Element, TextNode


class _NavigationStack:
    def __init__(self, initial_route: str) -> None:
        self._data = [initial_route]

    @property
    def top(self) -> str:
        return self._data[-1]

    def push(self, route: str) -> None:
        self._data.append(route)

    def pop(self) -> str:
        return self._data.pop()


class RouteManager:
    """A static class that enables navigation between various views in a PyFyre application."""

    _stack: _NavigationStack
    _routes_builder: Dict[str, Callable[..., Node]] = {}
    _routes: Dict[str, Optional[Node]] = {}
    _root_node = document.select_one("#root")

    @staticmethod
    def initialize(routes: Dict[str, Callable[..., Node]]) -> None:
        """:meta private:"""
        initial_route = RouteManager.parse_route(window.location.href)
        RouteManager._stack = _NavigationStack(initial_route)
        RouteManager._routes_builder = routes

        @window_event_listener("popstate")
        def onpopstate(event: DOMEvent) -> None:
            RouteManager.change_route(window.location.href)

    @staticmethod
    def parse_route(route_name: str) -> str:
        """Parse the ``route_name`` to turn it into a valid route name.

        Examples:
            | ``home`` -> ``/home``
            | ``contact/`` -> ``/contact``
            | ``about/this`` -> ``/about/this``
            | ``https://pyfyre.app/`` -> ``/``
            | ``https://pyfyre.app/about`` -> ``/about``
        """

        el = document.createElement("a")
        el.href = route_name
        route_name = str(el.pathname)

        if route_name == "/":
            return route_name

        return str(el.pathname).rstrip("/")

    @staticmethod
    def get_node(
        route_name: str,
        *,
        arg: Any = None,
        force_build: bool = True,
        parse_route: bool = True,
    ) -> Optional[Node]:
        """Call the corresponding route builder of the ``route_name``
        and return its ``Node``.

        Args:
            arg: The argument that will be passed in to the route builder.
            force_build: Whether to call the route builder even if it is already called before.
                By default, called route builders are cached.
            parse_route: Whether to call the
                ``RouteManager.parse_route`` method on the ``route_name``.

        Returns:
            The returned ``Node`` of the corresponding route builder of the ``route_name``.
            If the route doesn't exist, the default will be returned
            which has a 404 message.
        """

        if parse_route:
            route_name = RouteManager.parse_route(route_name)

        node = RouteManager._routes.get(route_name)

        if force_build or node is None:
            route_builder = RouteManager._routes_builder.get(route_name)

            if route_builder is not None:
                try:
                    node = route_builder(arg)
                except TypeError:
                    node = route_builder()
            else:
                node = None

            RouteManager._routes[route_name] = node

            if isinstance(node, Element):
                node.build_children()

        return node

    @staticmethod
    def render_route(
        route_name: str, *, arg: Any = None, force_build: bool = True
    ) -> None:
        """:meta private:"""
        node = RouteManager.get_node(
            route_name, arg=arg, force_build=force_build
        ) or TextNode("No route to render.")
        RouteManager._root_node.clear()
        RouteManager._root_node.attach(node.dom)

    @staticmethod
    def _update_page(
        route_data: Dict[str, Any], prev_route_data: Dict[str, Any]
    ) -> bool:
        if (route_data.get("head") or []) != (prev_route_data.get("head") or []):
            return False

        document.title = route_data.get("title")

        icon = document.select_one("link[rel~='icon']")
        if icon is not None:
            icon.href = route_data.get("icon")

        return True

    @staticmethod
    def change_route(
        route_name: str, *, arg: Any = None, force_build: bool = True
    ) -> None:
        """:meta private:"""
        prev_route = RouteManager._stack.top

        if ROUTES.get(prev_route) is None:
            window.location.reload()
        else:
            route_name = RouteManager.parse_route(route_name)
            RouteManager._stack.push(route_name)
            route_data = ROUTES.get(route_name)

            if route_data is None:
                window.location.reload()
            else:
                has_same_head = RouteManager._update_page(
                    route_data, ROUTES[prev_route]
                )
                if has_same_head:
                    RouteManager.render_route(
                        route_name, arg=arg, force_build=force_build
                    )
                else:
                    window.location.reload()



"""Entry point for PyFyre CLI installed via pip or setuptools.

Example:
    >>> pyfyre create hello-world
"""

import os
import sys
import pathlib
from typing import List, Optional

_HELP_MESSAGE = (
    "Commands:\n"
    "    pyfyre help\n"
    "        Show this message.\n"
    "    pyfyre version\n"
    "        Show the current version of PyFyre.\n"
    "    pyfyre create [path]\n"
    "        Create a new PyFyre project in the specified path.\n"
    "        If no path is supplied, the default is 'pyfyre-app'.\n"
    "        The directory name of the path will be your project's name.\n"
    "    pyfyre run [port] [hostname]\n"
    "        Run the PyFyre project in the current directory in development mode.\n"
    "        If no port is supplied, the default is '5500'.\n"
    "        If no hostname is supplied, the default is 'localhost'.\n"
    "    pyfyre build\n"
    "        Build the PyFyre project in the current directory for production.\n"
)


def execute(args: Optional[List[str]] = None) -> None:
    args = args or sys.argv
    args_list: List[Optional[str]] = [None, None, None, None]

    for index, arg in enumerate(args):
        try:
            args_list[index] = arg
        except IndexError:
            break

    command = args_list[1] or "help"

    if command == "help":
        print(_HELP_MESSAGE)
    elif command == "version":
        print("PyFyre 0.6.6-alpha")
    elif command == "create":
        from pyfyre_cli.create_app import create_app

        app_name = args_list[2] or "pyfyre-app"
        app_path = pathlib.Path(os.path.abspath(app_name))
        app_path_parts = app_path.parts[:-1]

        if not app_path_parts:
            print("You can't create a project in that directory.")
        else:
            create_app(app_path.parts[-1], os.path.join(*app_path_parts))
    elif command == "run":
        from pyfyre_cli.run_app import run_app

        port = args_list[2] or "5500"
        hostname = args_list[3] or "localhost"

        run_app(port, hostname)
    elif command == "build":
        from pyfyre_cli.build_app import build_app

        build_app(production=True)
    else:
        print("Command not found.")
        print(_HELP_MESSAGE)



"""
This module contains the tools for building the app
for development and production purposes.

The build process includes but not limited to bundling of the files
inside the `src` directory as a Brython package to make it usable for the web.

All the build files are saved in the `build` directory so you can just
serve or deploy the `build` directory to the web.
"""

import os
import sys
import shutil
import pathlib
import warnings
import importlib
import subprocess
from typing import List
from pyfyre_cli.utils import in_path

try:
    sys.path.append(os.getcwd())
    import settings
except ModuleNotFoundError:
    print("This directory is not a PyFyre project.")
    exit()


def _generate_page_head(*, production: bool) -> List[str]:
    head: List[str] = []

    if production:
        head.append('<script src="/modules.brython.js" defer="defer"></script>')
    else:
        head.append(
            '<script src="https://cdnjs.cloudflare.com/ajax/libs/brython/'
            '3.10.7/brython_stdlib.min.js" defer="defer"></script>'
        )

    if settings.PYTHON_DEPENDENCIES:
        head.append(
            '<script src="/cpython_packages.brython.js" defer="defer"></script>'
        )

    return head


def _generate_page_body(route_name: str) -> str:
    shutil.copytree("src", "__temp__", dirs_exist_ok=True)
    shutil.copy("settings.py", "__temp__")

    with in_path("__temp__", append_to_sys_path=True):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            import pyfyre

            pyfyre.PRODUCTION = True

            import index
            from pyfyre.router import RouteManager

            # Set [parse_route] to `False` because the [get_node] method
            # uses a Brython function which will not work
            # if [parse_route] is `True` in this context because
            # we are running the app using Python and not Brython.
            # Brython functions like [createElement] will not work in this context
            # that's why we would not want to parse the route which uses
            # a Brython function (document.createElement).
            html = RouteManager.get_node(route_name, parse_route=False)
            html = html.html() if html else "No route to render."

    shutil.rmtree("__temp__")
    return html


def create_pages(*, production: bool) -> None:
    importlib.reload(settings)

    with open("template.html") as file:
        html_template = file.read()

    for route, data in settings.ROUTES.items():
        directory = os.path.join(
            "build" if production else "_pyfyre", *route.split("/")
        )
        pathlib.Path(directory).mkdir(parents=True, exist_ok=True)

        with open(os.path.join(directory, "index.html"), "w") as file:
            head = _generate_page_head(production=production)
            html = html_template.format(
                prod_env=production,
                brython_options=str(
                    {
                        "debug": 0 if production else 1,
                        "cache": 1 if production else 0,
                    }
                ),
                title=data.get("title", "A PyFyre App"),
                icon=data.get("icon", ""),
                head="\n\t\t".join(head + data.get("head", [])),
                body=_generate_page_body(route) if production else "",
            )
            file.write(html)


def _cherry_pick_modules() -> None:
    shutil.copy(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "copybin", "brython_stdlib.js"
        ),
        "__temp__",
    )

    with in_path("__temp__"):
        # This command will detect the only Python modules that are used
        # in the project and will generate a cherry-picked standard libraries
        # for the project to run instead of using bloated `brython_stdlib.js`
        subprocess.run(["brython-cli", "make_modules"])
        os.rename("brython_modules.js", "modules.brython.js")
        shutil.copy("modules.brython.js", os.path.join("..", "build"))

        # Remove all files and folders except `index.py` to avoid duplicate VFS
        for file in os.listdir():
            if file == "index.py":
                continue

            if os.path.isfile(file):
                os.remove(file)
            else:
                shutil.rmtree(file)


def bundle_scripts(*, production: bool) -> None:
    shutil.copytree("src", "__temp__", dirs_exist_ok=True)
    shutil.copy("settings.py", "__temp__")

    temp_dir = os.path.abspath("__temp__")
    with in_path(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")):
        shutil.copytree("pyfyre", os.path.join(temp_dir, "pyfyre"), dirs_exist_ok=True)

    if production:
        subprocess.run(
            ["pyminify", "__temp__", "--in-place", "--remove-literal-statements"]
        )

        subprocess.run(
            [
                "autoflake",
                "__temp__",
                "-r",
                "--in-place",
                "--quiet",
                "--remove-unused-variables",
                "--remove-all-unused-imports",
                "--remove-duplicate-keys",
            ]
        )

        _cherry_pick_modules()

    with in_path("__temp__"):
        subprocess.run(["brython-cli", "make_package", "src"])
        shutil.copy(
            "src.brython.js", os.path.join("..", "build" if production else "_pyfyre")
        )

    shutil.rmtree("__temp__")


def add_cpython_packages(*, production: bool) -> None:
    importlib.reload(settings)

    for package_name in settings.PYTHON_DEPENDENCIES:
        subprocess.run(["brython-cli", "add_package", package_name])

    packages_dir = os.path.join("Lib", "site-packages")

    if not os.path.isdir(packages_dir):
        return

    with in_path(packages_dir):
        subprocess.run(["brython-cli", "make_package", "cpython_packages"])
        shutil.copy(
            "cpython_packages.brython.js",
            os.path.join("..", "..", "build" if production else "_pyfyre"),
        )

    shutil.rmtree("Lib")


def build_app(*, production: bool = False) -> None:
    os.environ["PYTHONUTF8"] = "1"

    if production:
        print("Building app...")

    # Mirror the public files of the user to the build files
    if os.path.isdir("public"):
        shutil.copytree(
            "public", "build" if production else "_pyfyre", dirs_exist_ok=True
        )

    create_pages(production=production)
    bundle_scripts(production=production)
    add_cpython_packages(production=production)

    if production:
        print("App successfully built.")



import traceback
from typing import Any, Type
from types import TracebackType
from pyfyre.styles import Style
from pyfyre.nodes import Element, Text


class DebugError(Element):
    """An error message with traceback.
    Typically used for debugging on development environment.

    Args:
        exc_type: Type of the exception.
        exc_value: The exception itself.
        exc_traceback: Traceback object which typically encapsulates the
            call stack at the point where the exception last occurred.
    """

    def __init__(
        self,
        exc_type: Type[Exception],
        exc_value: Exception,
        exc_traceback: TracebackType,
    ) -> None:
        tr = traceback.format_exc()
        super().__init__(
            "div",
            lambda: [
                Element(
                    "p",
                    lambda: [Text("Unhandled Runtime Error")],
                    styles=[
                        Style(
                            color="black",
                            font_size="1.5rem",
                            font_weight="bold",
                            margin_top="0",
                            margin_bottom="7px",
                        )
                    ],
                ),
                Element(
                    "span",
                    lambda: [Text(f"{exc_type.__name__}: {exc_value}")],
                    styles=[
                        Style(
                            color="#ff3131",
                            font_weight="bold",
                            font_family="monospace",
                            font_size="0.91rem",
                        )
                    ],
                ),
                Element(
                    "p",
                    lambda: [Text("Traceback")],
                    styles=[
                        Style(
                            color="black",
                            font_size="1.3rem",
                            font_weight="bold",
                            margin_bottom="7px",
                        )
                    ],
                ),
                Element(
                    "p",
                    lambda: [Element("pre", lambda: [Text(tr)])],
                    styles=[
                        Style(
                            background_color="black",
                            color="white",
                            padding="1px 15px",
                            font_size="0.9rem",
                            font_weight="normal",
                        )
                    ],
                ),
            ],
            styles=[
                Style(
                    padding="15px",
                    background_color="white",
                    box_shadow="0 0 10px #888888",
                    border_top="5px solid #ff726f",
                    border_radius="5px",
                    font_family="Arial",
                    display="inline-block",
                )
            ],
        )


class ErrorMessage(Element):
    """A simple error message. Typically used on production environment."""

    def __init__(self, message: Any) -> None:
        super().__init__(
            "div",
            lambda: [
                Element(
                    "p",
                    lambda: [Text(message)],
                    styles=[
                        Style(
                            color="black",
                            font_size="1rem",
                            margin="0",
                            font_weight="bold",
                        )
                    ],
                )
            ],
            styles=[
                Style(
                    padding="15px",
                    background_color="white",
                    box_shadow="0 0 10px #888888",
                    border_top="5px solid #ff726f",
                    border_radius="5px",
                    font_family="Arial",
                    display="inline-block",
                )
            ],
        )



from .base import Node, Element, E, TextNode, Text
from .buttons import Button
from .futures import FutureElement
from .inputs import TextInput
from .links import Link, RouterLink
from .lists import ListBuilder
from .widgets import Widget, FutureWidget

__all__ = [
    "Node",
    "Element",
    "E",
    "TextNode",
    "Text",
    "Button",
    "FutureElement",
    "TextInput",
    "Link",
    "RouterLink",
    "ListBuilder",
    "Widget",
    "FutureWidget",
]



from pyfyre.nodes import Node
from typing import Dict, Callable
from browser import window, DOMEvent
from pyfyre.router import RouteManager
from pyfyre.events import PyFyreEventType
from pyfyre.exceptions import PyFyreException
from .states import State, EventState, MediaQuery
from .styles import Style

__all__ = [
    "PRODUCTION",
    "render",
    "State",
    "EventState",
    "MediaQuery",
    "Style",
]

_rendered = False
PRODUCTION: bool = False
"""Whether the app is running on production environment or not."""


def _set_custom_window_events() -> Dict[str, DOMEvent]:
    events = {}

    for pyfyre_event in PyFyreEventType:
        event = window.Event.new("Event")
        event.initEvent(pyfyre_event.value, True, True)
        events[pyfyre_event.value] = event

    return events


def render(routes: Dict[str, Callable[..., Node]]) -> None:
    """The main function for initializing a PyFyre application.
    This function can only be called once.

    Args:
        routes: Pairs of route name and its corresponding route builder.
            The route builder returns a ``Node`` that will be rendered on the web page.
            The route builder optionally accepts a single argument.
            The value of the argument may be passed in by a ``RouterLink``.

    Raises:
        PyFyreException: When this function is called again.

    **Example:**

    .. code-block:: python

        RouterLink("/random", arg="A Random Title")

        render({
            "/": lambda: HomePage(),
            "/about": lambda: AboutPage(),
            "/contact": lambda: ContactPage(),
            "/random": lambda title: RandomPage(title)  # The argument defaults to None.
        })
    """

    global _rendered
    if _rendered:
        raise PyFyreException("'render' function can only be called once")
    _rendered = True

    custom_events = _set_custom_window_events()

    RouteManager.initialize(routes)
    RouteManager.render_route(window.location.pathname)
    window.dispatchEvent(custom_events["pyfyreload"])



from browser import DOMEvent
from pyfyre.styles import Style
from pyfyre.states import BaseState
from pyfyre.nodes.base import Node, Element
from typing import Callable, Optional, Dict, List


class Button(Element):
    """Represents an HTML ``<button>``.

    Args:
        onclick: Called when this button is clicked.
    """

    def __init__(
        self,
        onclick: Callable[[DOMEvent], None],
        children: Optional[Callable[[], List[Node]]] = None,
        *,
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        super().__init__("button", children, styles=styles, states=states, attrs=attrs)
        self.add_event_listener("click", onclick)



import sys
import pyfyre
from typing import Type
from browser import document
from types import TracebackType
from pyfyre.styles import Style
from pyfyre.states import BaseState, State
from abc import ABC, abstractmethod
from browser import DOMNode, DOMEvent
from typing import Any, Dict, List, Optional, Callable, Union


class Node(ABC):
    """Represents an HTML DOM node.

    Attributes:
        dom (DOMNode): Brython ``DOMNode`` type.
            The corresponding HTML DOM node of this object.
    """

    def __init__(self) -> None:
        self.dom: DOMNode = self.create_dom()

    @abstractmethod
    def create_dom(self) -> DOMNode:
        raise NotImplementedError

    @abstractmethod
    def update_dom(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def html(self) -> str:
        """HTML representation of this object."""
        raise NotImplementedError


class Element(Node):
    """Represents an HTML DOM element.

    Args:
        tag_name: HTML tag name of this element.
        children: Builder for the children nodes of this element.
        styles: CSS styling of this element.
            This will be combined into a single ``Style`` object.
            The styles are evaluated in order from left to right.
        states: State dependencies of this object.
            If one of the state dependencies has changed its state,
            the ``update_dom`` method of this object will be called.
        attrs: HTML attributes of this element.

    Attributes:
        tag_name (str): HTML tag name of this element.
        children (List[Node]): Children nodes of this element.
            This is equal to the return of the ``children`` argument.
        style (~pyfyre.Style): CSS styling of this element.
        states (List[~pyfyre.BaseState]): State dependencies of this object.
            If one of the state dependencies has changed its state,
            the ``update_dom`` method of this object will be called.
        attrs (Dict[str, str]): HTML attributes of this element.
    """

    def __init__(
        self,
        tag_name: str,
        children: Optional[Callable[[], List[Node]]] = None,
        *,
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None,
    ) -> None:
        self.tag_name = tag_name
        self.children: List[Node] = []
        self.style = Style.from_styles(styles) if styles else Style()
        self.states = states or []
        self.attrs = attrs or {}

        def children_builder() -> List[Node]:
            if children is None:
                return self.children

            try:
                return children()
            except Exception:
                return self.on_build_error(*sys.exc_info())

        self._children_builder = children_builder

        def update_style_attr() -> None:
            self.attrs = attrs or {}

            if "style" in self.attrs:
                self.attrs["style"] += f"; {self.style.css()}"
            else:
                self.attrs["style"] = self.style.css()

            if getattr(self, "dom", None) is not None:
                self.dom.setAttribute("style", self.attrs["style"])

        self.style.add_listener(update_style_attr)
        if self.style.props:
            update_style_attr()

        for state in self.states:
            state.add_listener(self.update_dom)

        super().__init__()

    def on_build_error(
        self,
        exc_type: Type[Exception],
        exc_value: Exception,
        exc_traceback: TracebackType,
    ) -> List[Node]:
        """:meta private:"""
        if pyfyre.PRODUCTION:
            from pyfyre.presets.errors import ErrorMessage

            return [ErrorMessage("An error occurred while loading this element")]

        from pyfyre.presets.errors import DebugError

        return [
            DebugError(
                exc_type=exc_type,
                exc_value=exc_value,
                exc_traceback=exc_traceback,
            )
        ]

    def build_children(self, *, propagate: bool = True) -> None:
        """:meta private:"""
        self.dom.clear()
        self.children = self._children_builder()

        for child in self.children:
            self.dom.attach(child.dom)

            if propagate and isinstance(child, Element):
                child.build_children()

    def create_dom(self) -> DOMNode:
        """Create a new HTML DOM element from this object.

        Returns:
            Brython ``DOMNode`` type.
        """
        el = document.createElement(self.tag_name)

        for attr_name, attr_value in self.attrs.items():
            el.setAttribute(attr_name, attr_value)

        return el

    def update_dom(self) -> None:
        """Update the corresponding HTML DOM element of this object.
        This rebuilds the children of this element recursively.
        """
        self.update_attrs()
        self.build_children(propagate=False)

        for child in self.children:
            child.update_dom()

    def update_attrs(self) -> None:
        """Update the attributes of the corresponding HTML DOM element of this object."""
        for attr_name, attr_value in self.attrs.items():
            self.dom.setAttribute(attr_name, attr_value)

    def html(self) -> str:
        def attributes() -> str:
            attrs = [f'{name}="{value}"' for name, value in self.attrs.items()]
            return " " + " ".join(attrs) if attrs else ""

        result = f"<{self.tag_name}{attributes()}>"

        for child in self.children:
            result += child.html()

        return result + f"</{self.tag_name}>"

    def add_event_listener(
        self, event_type: str, callback: Callable[[DOMEvent], None]
    ) -> None:
        """Calls the ``callback`` whenever the specified ``event_type``
        is delivered to this element.
        """
        self.dom.bind(event_type, callback)


class TextNode(Node):
    """Represents an HTML DOM text node.

    Args:
        *values: Value of the text.
            You can pass in a ``State`` object to automatically update the value of the text
            when the state dependency has changed its state.
    """

    def __init__(self, *values: Union[State[Any], Any]) -> None:
        self._value = self._process_values(*values)
        super().__init__()

    def _process_values(self, *values: Union[State[Any], Any]) -> str:
        def add_values() -> str:
            result = ""

            for value in values:
                if isinstance(value, State):
                    result += str(value.value)
                else:
                    result += str(value)

            return result

        def update_value() -> None:
            self._value = add_values()
            self.update_dom()

        for value in values:
            if isinstance(value, State):
                value.add_listener(update_value)

        return add_values()

    @property
    def value(self) -> str:
        """Value of the text."""
        self._value = self.dom.nodeValue
        return self._value

    def create_dom(self) -> DOMNode:
        """Create a new HTML DOM text node from this object.

        Returns:
            Brython ``DOMNode`` type.
        """
        return document.createTextNode(self._value)

    def update_dom(self) -> None:
        """:meta private:"""
        self.dom.nodeValue = self._value

    def html(self) -> str:
        return self._value


E = Element
Text = TextNode



class PyFyreException(Exception):
    pass


class FutureElementException(PyFyreException):
    pass


class FutureNoResult(FutureElementException):
    pass


class FutureCancelled(FutureElementException):
    pass


class FutureAlreadyDone(FutureElementException):
    pass



from pyfyre.nodes import Element


class Loading(Element):
    """A loading animation.

    Source: https://loading.io/css/
    """

    def __init__(self) -> None:
        super().__init__(
            "div", lambda: [Element("div") for _ in range(12)], attrs={"class": "lds"}
        )



import os
from typing import List
from setuptools import setup


def get_readme() -> str:
    with open("README.md") as file:
        return file.read()


def get_package_files(directory: str) -> List[str]:
    paths = []

    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))

    return paths


packages = ["pyfyre", "pyfyre_cli", "browser", "interpreter", "javascript"]

setup(
    name="pyfyre",
    version="0.6.6",
    description="A fast, declarative, and incrementally adoptable Python web "
    "frontend framework for building reactive web user interfaces.",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    author="StackSpace",
    author_email="stackspace.ph@gmail.com",
    url="https://pyfyre.netlify.app/",
    download_url="https://pypi.org/project/pyfyre/",
    packages=packages,
    package_data={
        "": sum([get_package_files(directory) for directory in packages], [])
    },
    license="MIT",
    project_urls={
        "Source": "https://github.com/pyfyre/pyfyre",
        "Updates": "https://www.facebook.com/pyfyreframework/",
        "Documentation": "https://pyfyre-docs.netlify.app/",
        "Community": "https://discord.gg/YzEDuqhgZJ",
    },
    entry_points={"console_scripts": ["pyfyre = pyfyre_cli:execute"]},
    install_requires=[i.strip() for i in open("requirements.txt").readlines()],
)



from typing import Any


class EmptyObject:
    def __init__(self, class_name: str) -> None:
        self.class_name = class_name
        self.__qualname__ = class_name

    def __getattr__(self, attr: str) -> Any:
        return EmptyObject(attr)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return EmptyObject("")

    def __getitem__(self, key: str) -> Any:
        return EmptyObject(key)

    def __str__(self) -> str:
        return self.class_name

    def __add__(self, other: str) -> str:
        return str(self) + other

    def __radd__(self, other: str) -> str:
        return other + str(self)


Interpreter: Any = EmptyObject("Interpreter")
Inspector: Any = EmptyObject("Inspector")



from pyfyre.styles import Style
from pyfyre.states import BaseState
from pyfyre.nodes.base import Node, Element
from browser import DOMEvent, window, document
from typing import Any, Optional, Dict, List, Callable


class Link(Element):
    """Represents an HTML ``<a>``.

    Args:
        href: Hyperlink of the page the link goes to.

    Attributes:
        href (str): Hyperlink of the page the link goes to.
    """

    def __init__(
        self,
        href: str,
        children: Optional[Callable[[], List[Node]]] = None,
        *,
        styles: Optional[List[Style]] = None,
        states: Optional[List[BaseState]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        self.href = href
        attrs = attrs or {}
        attrs["href"] = href
        super().__init__("a", children, styles=styles, states=states, attrs=attrs)

    @property
    def url(self) -> str:
        """URL of the page the link goes to."""
        el = document.createElement("a")
        el.href = self.href
        return el.href

    def is_internal(self) -> bool:
        """Whether the href is linked within the same domain."""
        el = document.createElement("a")
        el.href = self.href
        return bool(el.host == window.location.host)


class RouterLink(Link):
    """Navigates to a different route inside the website as a single-page application.

    Args:
        arg: The argument that will be passed in to the route builder.
        force_build: Whether to call the route builder even if it is already called before.
            By default, called route builders are cached.
    """

    def __init__(
        self,
        href: str,
        children: Optional[Callable[[], List[Node]]] = None,
        *,
        arg: Any = None,
        force_build: bool = True,
        styles: Optional[List[Style]] = None,
        attrs: Optional[Dict[str, str]] = None
    ) -> None:
        super().__init__(href, children, styles=styles, attrs=attrs)

        def change_route(event: DOMEvent) -> None:
            # Import here due to cicular import problem
            from pyfyre.router import RouteManager

            window.history.pushState(None, None, self.url)
            event.preventDefault()
            RouteManager.change_route(href, arg=arg, force_build=force_build)

        self.add_event_listener("click", change_route)



from typing import List, Callable


class EventMixin:
    """A mixin class which implements event-like functionalities
    to the class inheriting this.
    """

    def __init__(self) -> None:
        self._listeners: List[Callable[[], None]] = []

    def call_listeners(self) -> None:
        """Call all the listeners of this event."""
        for listener in self._listeners:
            listener()

    def add_listener(self, listener: Callable[[], None]) -> None:
        """Add the ``listener`` to this event."""
        self._listeners.append(listener)

    def remove_listener(self, listener: Callable[[], None]) -> int:
        """Remove the ``listener`` to this event.

        Returns:
            The number of listeners removed, which is typically 1,
            unless the ``listener`` was added more than once.
        """
        remaining_listeners = []
        removed = 0

        for c in self._listeners:
            if c == listener:
                removed += 1
            else:
                remaining_listeners.append(c)

        self._listeners = remaining_listeners
        return removed
